setwd("~/rsch/OutDepSamp/Projects/Software/linear.longit.acl/")
source("SimGenDatFns.R")
N          <- 750
RanEffDist <- "GaussianBal"

nlow       <-  10
nhigh      <-  10

ProfileCol <- 5
Beta       <- c(5, 1, -2.5, 0.75, 0)
sig.b0     <- 5
sig.b1     <- 1.25
sig.e      <- 5
rho        <- -0.25
prob.conf  <- 0.5
prob.grp.0 <- 0.40
prob.grp.1 <- 0.60
num.in.ods <- 250
inits        <- c(Beta, log(sig.b0), log(sig.b1), log((1+rho)/(1-rho)), log(sig.e))

##################### Calculate population quantiles used for for cutoffs
PopnQuants <- est.quants(N=25000, n=nhigh, beta=Beta, sig.b0=sig.b0, sig.b1=sig.b1, 
                             rho=rho, sig.e=sig.e, pr.grp.0=prob.grp.0, pr.grp.1=prob.grp.1, pr.conf=prob.conf,
                             quant=c(.05, .1,.12, .2, .8, .88, .9, .95), n.low=nlow, n.high=nhigh, dist=RanEffDist)
                             
PopnQuants.bivar <- est.bivar.lims(N=25000, n=nhigh, beta=Beta, sig.b0=sig.b0, sig.b1=sig.b1, rho=rho, 
                                   sig.e=sig.e, pr.grp.0=prob.grp.0, pr.grp.1=prob.grp.1, pr.conf=prob.conf,
                             quant=c(.6,.76), n.low=nlow, n.high=nhigh, dist=RanEffDist)
#####################
set.seed(100)
dat              <- GenPopnData(N=N, n=nhigh, Beta, sig.b0, sig.b1, rho, sig.e, 
                                pr.grp.0=prob.grp.0, pr.grp.1=prob.grp.1, pr.conf=prob.conf,
                                n.low=nlow, n.high=nhigh, dist=RanEffDist)
ods.int       <- ODS.Sampling(dat=dat, PopnQuants=PopnQuants, w.function="intercept", quants=c(.12,.88), 
                              TargetNSampledPerStratum=round(num.in.ods*c(9/25, 7/25, 9/25)), SamplingStrategy="IndepODS") 
ods.slp       <- ODS.Sampling(dat=dat, PopnQuants=PopnQuants, w.function="slope", quants=c(.12,.88), 
                              TargetNSampledPerStratum=round(num.in.ods*c(9/25, 7/25, 9/25)), SamplingStrategy="IndepODS") 
ods.biv       <- ODS.Sampling.Bivar(dat=dat, PopnQuants=PopnQuants.bivar, PopnPropInRectangle=.76,                                            
                                    TargetNSampledPerStratum=round(num.in.ods*c(7/25, 18/25)), SamplingStrategy="IndepODS")

#########################
library(linear.longit.acl)

odsInt <- data.frame(cbind(ods.int$id[ods.int$InSample], ods.int$Y[ods.int$InSample], ods.int$X[ods.int$InSample,-1] ))
names(odsInt) <- c("id", "Y", "time","snp","snptime","confounder")
odsSlp <- data.frame(cbind(ods.slp$id[ods.slp$InSample], ods.slp$Y[ods.slp$InSample], ods.slp$X[ods.slp$InSample,-1] ))
names(odsSlp) <- c("id", "Y", "time","snp","snptime","confounder")
odsBiv <- data.frame(cbind(ods.biv$id[ods.biv$InSample], ods.biv$Y[ods.biv$InSample], ods.biv$X[ods.biv$InSample,-1] ))
names(odsBiv) <- c("id", "Y", "time","snp","snptime","confounder")
write.csv(odsInt, "odsInt.csv")
write.csv(odsSlp, "odsSlp.csv")
write.csv(odsBiv, "odsBiv.csv")


## Run the code for one realization from simulation scenarion e) in Schildcrout et al (2015) with the exception that we 
## estimate the parameter associated with C_i here.  In the paper it was not estimated.  
##
## Expect Fit.Biv to take substantially longer to fit than the others
odsInt <- read.csv("odsInt.csv")
odsSlp <- read.csv("odsSlp.csv")
odsBiv <- read.csv("odsSlp.csv")

Fit.Int  <- acml.linear(y=odsInt$Y, 
                        x=as.matrix(cbind(1, odsInt[,c("time","snp","snptime","confounder")])), 
                        z=as.matrix(cbind(1, odsInt$time)), 
                        id=odsInt$id, 
                        InitVals=c(5,  1, -2.5,  0.75,  0,  1.6094379,  0.2231436, -0.5108256,  1.6094379), 
                        ProfileCol=NA, 
                        cutpoints=c(-2.569621, 9.992718), 
                        SampProb=c(1, 0.1228, 1), 
                        w.function="intercept")

Fit.Slp  <- acml.linear(y=odsSlp$Y, 
                        x=as.matrix(cbind(1, odsSlp[,c("time","snp","snptime","confounder")])), 
                        z=as.matrix(cbind(1, odsSlp$time)), 
                        id=odsSlp$id, 
                        InitVals=c(5,  1, -2.5,  0.75,  0,  1.6094379,  0.2231436, -0.5108256,  1.6094379), 
                        ProfileCol=NA, 
                        cutpoints=c(-0.7488912,  3.4557775), 
                        SampProb=c(1, 0.1228, 1), 
                        w.function="intercept")

Fit.Biv  <- acml.linear(y=odsBiv$Y, 
                        x=as.matrix(cbind(1, odsBiv[,c("time","snp","snptime","confounder")])), 
                        z=as.matrix(cbind(1, odsBiv$time)), 
                        id=odsBiv$id, 
                        InitVals=c(5,  1, -2.5,  0.75,  0,  1.6094379,  0.2231436, -0.5108256,  1.6094379), 
                        ProfileCol=NA, 
                        cutpoints=c(-4.413225, 11.935188, -1.390172, 4.084768), 
                        SampProb=c(0.122807, 1), 
                        w.function="bivar")
Fit.Int
Fit.Slp
Fit.Biv

################ Fitting the intercept based sampling model. Profiling out the fifth column of X
w.function <- ods.int$w.function
SampProb   <- ods.int$SampProb
CutPoint   <- ods.int$cutpoint

y.insample   <- ods.int$Y[ods.int$InSample]
x.insample   <- ods.int$X[ods.int$InSample,]
z.insample   <- ods.int$Z[ods.int$InSample,]
id.insample  <- ods.int$id[ods.int$InSample]
        
date1 <- date()
    Fit.int  <- acml.linear(y=y.insample, x=x.insample, z=z.insample, id=id.insample, InitVals=inits, ProfileCol=ProfileCol, 
                      cutpoints=CutPoint, SampProb=SampProb, w.function=w.function)
date()                 

################ Fitting the slope-based sampling model. Profiling out the fifth column of X
w.function <- ods.slp$w.function
SampProb   <- ods.slp$SampProb
CutPoint   <- ods.slp$cutpoint

y.insample   <- ods.slp$Y[ods.slp$InSample]
x.insample   <- ods.slp$X[ods.slp$InSample,]
z.insample   <- ods.slp$Z[ods.slp$InSample,]
id.insample  <- ods.slp$id[ods.slp$InSample]
        
date1 <- date()
    Fit.slp  <- acml.linear(y=y.insample, x=x.insample, z=z.insample, id=id.insample, InitVals=inits, ProfileCol=ProfileCol, 
                      cutpoints=CutPoint, SampProb=SampProb, w.function=w.function)
date()       

################ Fitting the (intercept,slope)-based sampling model. Profiling out the fifth column of X

w.function <- ods.biv$w.function
SampProb   <- ods.biv$SampProb
CutPoint   <- ods.biv$cutpoint

y.insample   <- ods.biv$Y[ods.biv$InSample]
x.insample   <- ods.biv$X[ods.biv$InSample,]
z.insample   <- ods.biv$Z[ods.biv$InSample,]
id.insample  <- ods.biv$id[ods.biv$InSample]
        
date1 <- date()
    Fit.biv  <- acml.linear(y=y.insample, x=x.insample, z=z.insample, id=id.insample, InitVals=inits, ProfileCol=ProfileCol, 
                      cutpoints=CutPoint, SampProb=SampProb, w.function=w.function)
date()    
