#library(MASS, lib.loc="/usr/local/biostat_r/lib/R")
library(MASS)

## Standardized Gamma Random effect
gen.std.gam <- function(n.obs, shape, rate){
    (rgamma(n.obs, shape, rate) - shape/rate)/(sqrt(shape/(rate^2)))
}

## Standardized T random effect
gen.std.t <- function(n.obs, df){
	rt(n.obs, df)*sqrt((df-2)/df) 
	}

## Standardized binormal random effect
gen.std.binormal <- function(n.obs, mn0, mn1, sd0, sd1, p1){
	group <- rbinom(n.obs, 1, p1)
	val <- rnorm(n.obs, mn0, sd0)*(1-group)+rnorm(n.obs, mn1, sd1)*group
	grp.tmp <- rbinom(50000, 1, p1)
	val.tmp <- rnorm(50000, mn0, sd0)*(1-grp.tmp)+rnorm(50000, mn1, sd1)*grp.tmp
	
	out <- (val-mean(val.tmp))/(sqrt(var(val.tmp)))
    out	
	}
	
## Standardized mixture of normals with difference variances
gen.std.mixnorm <- function(n.obs, mn0, mn1, sd0, sd1, p1, group){
	
	val <- rnorm(n.obs, mn0, sd0)*(1-group)+rnorm(n.obs, mn1, sd1)*group
	
	grp.tmp <- rbinom(50000, 1, p1)
	val.tmp <- rnorm(50000, mn0, sd0)*(1-grp.tmp)+rnorm(50000, mn1, sd1)*grp.tmp
	
	out <- (val-mean(val.tmp))/(sqrt(var(val.tmp)))
    out	
	}

GenPopnData <- function(N = 1000, n = 11, beta = c(1, 0.25, 0, 0.25),
                     sig.b0 = 0.25, sig.b1 = 0.25, rho = 0, sig.e = 0.5, 
                     pr.conf=0.25, pr.grp.0 = 0.1, pr.grp.1 = 0.2, n.low=11, n.high=11,
                     dist){
    id       <- rep(1:N, each=n)
    conf.i   <- rbinom(N, 1, pr.conf)
    pr.grp.i <- ifelse(conf.i==1, pr.grp.1, pr.grp.0)
    grp.i    <- rbinom(N,1,pr.grp.i)
    conf     <- rep(conf.i, each=n) 
    grp      <- rep(grp.i, each=n)
    #grp      <- rep(rbinom(N, 1, prob.grp), each=n)
    ###### Add between subject variation in the time variable
    ###### Note that we are creating ICC equal to 0.5
    ###### var(sample(seq(-2,2,length.out=10), 100000, replace=TRUE))

    bl.age   <- rep(rep(0,N), each=n)
    time     <- rep(seq(-2,2, length.out=n), N)
    age      <- bl.age + time
    pr.grp.1.overall <- pr.conf*pr.grp.1 + (1-pr.conf)*pr.grp.0

    cov.mat  <- matrix(c(sig.b0^2, rho*sig.b0*sig.b1, rho*sig.b0*sig.b1, sig.b1^2),2,2)
    if (dist %in% c("Gaussian","GaussianBal","Gamma5err","T5err","MixNorm3err")){bi <- mvrnorm(N, mu=c(0,0), Sigma=cov.mat)
    }else if (dist=="Gamma"){b0i <- gen.std.gam(n.obs=N, shape=3, rate=sqrt(3))*sig.b0
    	               b1i <- gen.std.gam(n.obs=N, shape=3, rate=sqrt(3))*sig.b1
    	               bi  <- cbind(b0i, b1i)
    }else if (dist=="Gamma5"){b0i <- gen.std.gam(n.obs=N, shape=5, rate=sqrt(3))*sig.b0
    	               b1i <- gen.std.gam(n.obs=N, shape=5, rate=sqrt(3))*sig.b1
    	               bi  <- cbind(b0i, b1i)
    }else if (dist=="Gamma10"){b0i <- gen.std.gam(n.obs=N, shape=10, rate=sqrt(3))*sig.b0
    	               b1i <- gen.std.gam(n.obs=N, shape=10, rate=sqrt(3))*sig.b1
    	               bi  <- cbind(b0i, b1i)
    }else if (dist=="Gamma15"){b0i <- gen.std.gam(n.obs=N, shape=15, rate=sqrt(3))*sig.b0
    	               b1i <- gen.std.gam(n.obs=N, shape=15, rate=sqrt(3))*sig.b1
    	               bi  <- cbind(b0i, b1i)
    }else if (dist=="T5"){b0i <- gen.std.t(n.obs=N, df=5)*sig.b0
    	           b1i <- gen.std.t(n.obs=N, df=5)*sig.b1
    	           bi  <- cbind(b0i, b1i)
    }else if (dist=="T10"){b0i <- gen.std.t(n.obs=N, df=10)*sig.b0
    	           b1i <- gen.std.t(n.obs=N, df=10)*sig.b1
    	           bi  <- cbind(b0i, b1i)
    }else if (dist=="T15"){b0i <- gen.std.t(n.obs=N, df=15)*sig.b0
    	           b1i <- gen.std.t(n.obs=N, df=15)*sig.b1
    	           bi  <- cbind(b0i, b1i)
    }else if (dist=="Binormal1"){b0i <- gen.std.binormal(n.obs=N, mn0=0, mn1=1, sd0=1, sd1=1, p1=.25)*sig.b0
    	                  b1i <- gen.std.binormal(n.obs=N, mn0=0, mn1=1, sd0=1, sd1=1, p1=.25)*sig.b1
    	                  bi  <- cbind(b0i, b1i)
    }else if (dist=="Binormal2"){b0i <- gen.std.binormal(n.obs=N, mn0=0, mn1=2, sd0=1, sd1=1, p1=.25)*sig.b0
    	                  b1i <- gen.std.binormal(n.obs=N, mn0=0, mn1=2, sd0=1, sd1=1, p1=.25)*sig.b1
    	                  bi  <- cbind(b0i, b1i)
    }else if (dist=="MixNorm1.25"){b0i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(1.25), p1=pr.grp.1.overall, group=grp.i)*sig.b0
    	                   b1i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(1.25), p1=pr.grp.1.overall, group=grp.i)*sig.b1
    	                   bi  <- cbind(b0i, b1i)
    }else if (dist=="MixNorm1.5"){b0i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(1.5), p1=pr.grp.1.overall, group=grp.i)*sig.b0
    	                   b1i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(1.5), p1=pr.grp.1.overall, group=grp.i)*sig.b1
    	                   bi  <- cbind(b0i, b1i)
    }else if (dist=="MixNorm2"){b0i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(2), p1=pr.grp.1.overall, group=grp.i)*sig.b0
    	                   b1i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(2), p1=pr.grp.1.overall, group=grp.i)*sig.b1
    	                   bi  <- cbind(b0i, b1i)
    }else if (dist=="MixNorm2.25"){b0i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(2.25), p1=pr.grp.1.overall, group=grp.i)*sig.b0
    	                   b1i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(2.25), p1=pr.grp.1.overall, group=grp.i)*sig.b1
    	                   bi  <- cbind(b0i, b1i)
    }else if (dist %in% c("MixNorm3","MixNorm3err3")){b0i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(3), p1=pr.grp.1.overall, group=grp.i)*sig.b0
    	                   b1i <- gen.std.mixnorm(n.obs=N, mn0=0, mn1=0, sd0=1, sd1=sqrt(3), p1=pr.grp.1.overall, group=grp.i)*sig.b1
    	                   bi  <- cbind(b0i, b1i)
    }
    b        <- cbind(rep(bi[,1], each=n),rep(bi[,2], each=n))
    
    if (dist == "Gamma5err"){ 
    	    error <- gen.std.gam(n.obs=N*n, shape=5, rate=sqrt(3))*sig.e
    	}else if (dist == "T5err"){ 
    	    error <- gen.std.t(n.obs=N*n, df=5)*sig.e
    	}else if (dist %in% c("MixNorm3err","MixNorm3err3")){ 
    		error <- gen.std.mixnorm(n.obs=N*n, mn0=0, mn1=0, sd0=1, sd1=sqrt(3), p1=pr.grp.1.overall, group=grp)*sig.e
     }else{ 
     	error    <- rnorm(N*n, 0, sig.e)
     }

   
    X <- as.matrix(cbind(1, age, grp, age*grp, conf))
    Z <- X[,c(1:2)]
    Y <- X %*% beta + Z[,1]*b[,1] + Z[,2]*b[,2] + error

    ## Induce MCAR dropout
    n.obs <- rep(sample(rep(c(n.low:n.high),2), N, replace=TRUE), each=n) 
    obs.num <- rep(c(1:n), N)
    X <- X[obs.num<= n.obs,]
    Z <- Z[obs.num<= n.obs,]
    Y <- Y[obs.num<= n.obs]
    id <- id[obs.num<= n.obs]
    list(id=id, X=X, Y=Y, Z=Z, ni=c(unlist(tapply(Y,id,length))),
         N=N, n=n, beta=beta,sig.b0=sig.b0, sig.b1=sig.b1, rho=rho, sig.e=sig.e, pr.grp.1=pr.grp.1, pr.grp.0=pr.grp.0)
}

LinRegFn <- function(data){ X <- cbind(1, data[,3])
	                       Xt <- t(X)
	                       Y <- data$Y
	                       solve(Xt %*% X) %*% Xt %*% Y}
	                          
est.quants <- function(N, n, beta, sig.b0, sig.b1, rho, sig.e, pr.grp.0, pr.grp.1, pr.conf, quant, n.low=11, n.high=11, dist){
    d        <- GenPopnData(N=N, n=n, beta=beta, sig.b0=sig.b0, sig.b1=sig.b1, rho=rho, sig.e=sig.e, 
                           pr.conf=pr.conf, pr.grp.0 = pr.grp.0, pr.grp.1 = pr.grp.1, dist=dist)
    data.tmp <- data.frame(id=d$id, Y=d$Y, X.time=d$X[,2])
    out      <- matrix(unlist(lapply(split(data.tmp, data.tmp$id), LinRegFn)), byrow=TRUE, ncol=2)
    out      <- cbind(c(tapply(data.tmp$Y, data.tmp$id, mean)), out)
    r        <- rbind( quant,
                       quantile(out[,1], probs=quant), 
                       quantile(out[,2], probs=quant),
                       quantile(out[,3], probs=quant))
    r}

## Do a search to find the quantiles that correspond to the central rectangle that contains 60 and 80 percent of 
## the subject specific intercepts and slopes.  This is not necessary if slope and intercepts are independent
## but with unequal followup they were positiviely correlated.  Searches for the smallest 'rectangle' defined
## by quantiles that contains 60 and 80 percent of the data
est.bivar.lims <- function(N, n, beta, sig.b0, sig.b1, rho, sig.e, pr.grp.0, pr.grp.1, pr.conf, quants, n.low=11, n.high=11, dist){
	   d        <- GenPopnData(N=N, n=n, beta=beta, sig.b0=sig.b0, sig.b1=sig.b1, rho=rho, sig.e=sig.e, 
                                pr.conf=pr.conf, pr.grp.0 = pr.grp.0, pr.grp.1 = pr.grp.1, dist=dist)
        #d        <- GenPopnData(N, n, beta, sig.b0, sig.b1, rho, sig.e, prob.grp, n.low=n.low, n.high=n.high)
        data.tmp <- data.frame(id=d$id, Y=d$Y, X.time=d$X[,2])
        out      <- matrix(unlist(lapply(split(data.tmp, data.tmp$id), LinRegFn)), byrow=TRUE, ncol=2)
        out      <- cbind(c(tapply(data.tmp$Y, data.tmp$id, mean)), out)
        print(cor(out[,2], out[,3]))
        
        q1 <- .99
        Del <- 1
        while (Del>0.001){ q1 <- q1-.00025
                           Del <- abs(mean(out[,2] > quantile(out[,2], probs=1-q1) & 
                                           out[,2] < quantile(out[,2], probs=q1) &
                                           out[,3] > quantile(out[,3], probs=1-q1) & 
                                           out[,3] < quantile(out[,3], probs=q1)) - quants[1])
                           #print(c(q1,Del))
                           q1}
        q2 <- .99
        Del <- 1
        while (Del>0.001){ q2 <- q2-.00025
                           Del <- abs(mean(out[,2] > quantile(out[,2], probs=1-q2) & 
                                           out[,2] < quantile(out[,2], probs=q2) &
                                           out[,3] > quantile(out[,3], probs=1-q2) & 
                                           out[,3] < quantile(out[,3], probs=q2)) - quants[2])
                           q2}
       
        rbind( c( quantile(out[,2], probs=1-q1), quantile(out[,2], probs=q1), quantile(out[,3], probs=1-q1), quantile(out[,3], probs=q1)),
               c( quantile(out[,2], probs=1-q2), quantile(out[,2], probs=q2), quantile(out[,3], probs=1-q2), quantile(out[,3], probs=q2)))

}

ODS.Sampling.Bivar <- function(dat,                      ## a list generated from the GenPopnData() function
                               PopnQuantsBivar,          ## a matrix from est.bivar.lims function
                               PopnPropInRectangle,      ## Proportion of subjects in the central rectangle
                               TargetNSampledPerStratum, ## Theoretical (and possibly observed) number sampled per stratum
                               SamplingStrategy){        ## Options are "IndepODS" and "DepODS"
         
    NCohort                    <- length(unique(dat$id))
    NStratumThry               <- round(NCohort*c(PopnPropInRectangle, (1-PopnPropInRectangle)))
    SampProbThry               <- TargetNSampledPerStratum / NStratumThry
    SampProbThry[1] <- ifelse(SampProbThry[1]>1, 1, SampProbThry[1])
    SampProbThry[2] <- ifelse(SampProbThry[2]>1, 1, SampProbThry[2])
    
    Lims <- (PopnPropInRectangle==.6)*PopnQuantsBivar[1,] + (PopnPropInRectangle==.76)*PopnQuantsBivar[2,] 
   
    uid <- unique(dat$id)
    ni  <- c(unlist(tapply(dat$id, dat$id, length)))  
    SampVar <- NULL  
    for(i in uid){ yi      <- dat$Y[dat$id==i]
                   xi      <- dat$X[dat$id==i,1:2]
                   SampVar <- rbind(SampVar, c(mean(yi), solve(t(xi)%*%xi) %*% t(xi) %*% yi))
    }
   print(sum(SampVar[,2]>Lims[1] & SampVar[,2]<Lims[2]))
   print(sum(SampVar[,3]>Lims[3] & SampVar[,3]<Lims[4]))
 
    SampStratum  <- ifelse(SampVar[,2]>Lims[1] & SampVar[,2]<Lims[2] & 
                           SampVar[,3]>Lims[3] & SampVar[,3]<Lims[4], 1,2) 
    print(table(SampStratum))

    NperStratum  <- unlist(tapply(uid, SampStratum, length))
  
    
    SampProbiThry <- ifelse(SampStratum==1, SampProbThry[1],
                     ifelse(SampStratum==2, SampProbThry[2], NA))
    
    
    Sampled     <- rbinom(length(SampProbiThry), 1, SampProbiThry)
    SampProbObs <- c(tapply(Sampled, SampStratum, mean))

    SampProbiObs  <- ifelse(SampStratum==1, SampProbObs[1],
                     ifelse(SampStratum==2, SampProbObs[2], NA))

    ## Independent Sampling
    if (SamplingStrategy=="IndepODS") InODS <- uid[ Sampled==1] 
        
    TheSample <- dat$id %in% InODS                     
    X.ods     <- dat$X[TheSample,]
    Y.ods     <- dat$Y[TheSample]
    Z.ods     <- dat$Z[TheSample,]
    id.ods    <- dat$id[TheSample]
    if (SamplingStrategy=="IndepODS"){
        SampProbi.ods <- rep(SampProbiThry, ni)[TheSample]
        SampProb.ods  <- SampProbThry
    }
    #if (SamplingStrategy=="DepODS"){
    #	   SampProbi.ods <- rep(SampProbiObs, ni)[TheSample]
    #	   SampProb.ods  <- SampProbObs
    #}
    SampStrat.ods <- SampStratum[InODS]
    Qi <- SampVar[InODS,2:3]
    dup.id <- duplicated(dat$id)
    dat.univariate <- dat$X[!dup.id,]
    dat.univ.ods   <- dat.univariate[InODS,]

    list(#X=X.ods, Y=Y.ods, Z=Z.ods, id=id.ods,
    X=dat$X, Y=dat$Y, Z=dat$Z, id=dat$id,
      SampProb=SampProb.ods, SampProbi=SampProbi.ods,
      N=dat$N, n=dat$n, beta=dat$beta, sig.b0=dat$sig.b0, 
      sig.b1=dat$sig.b1, rho=dat$rho, sig.e=dat$sig.e,
      prob.grp=dat$prob.grp,w.function="bivar",
      #cutpoint=c(C1,C2), 
      SampStratum=SampStrat.ods, Qi=Qi, dat.univ.ods=dat.univ.ods,
      cutpoint=Lims, InSample=TheSample)
}

ODS.Sampling <- function(dat,                      ## a list generated from the GenPopnData() function
                         PopnQuants,               ## a matrix from est.quants function
                         w.function,               ## Response summary to sample on ("mean","intercept", or "slope")
                         quants,                   ## Population quantiles to define the theoretical sampling strata
                         TargetNSampledPerStratum, ## Theoretical (and possibly observed) number sampled per stratum
                         SamplingStrategy,         ## Options are "IndepODS" and "DepODS"
                         Univariate=FALSE){

    NCohort                    <- length(unique(dat$id))
    NStratumThry               <- round(NCohort*c(quants[1], quants[2]-quants[1], 1-quants[2]))
    SampProbThry               <- TargetNSampledPerStratum / NStratumThry
    SampProbThry[1] <- ifelse(SampProbThry[1]>1, 1, SampProbThry[1])
    SampProbThry[2] <- ifelse(SampProbThry[2]>1, 1, SampProbThry[2])
    SampProbThry[3] <- ifelse(SampProbThry[3]>1, 1, SampProbThry[3])
    
    C1 <- ifelse(w.function=="mean",      PopnQuants[2,match(quants[1], PopnQuants[1,])],
          ifelse(w.function=="intercept", PopnQuants[3,match(quants[1], PopnQuants[1,])],
          ifelse(w.function=="slope",     PopnQuants[4,match(quants[1], PopnQuants[1,])])))
    C2 <- ifelse(w.function=="mean",      PopnQuants[2,match(quants[2], PopnQuants[1,])],
          ifelse(w.function=="intercept", PopnQuants[3,match(quants[2], PopnQuants[1,])],
          ifelse(w.function=="slope",     PopnQuants[4,match(quants[2], PopnQuants[1,])])))
      
    uid <- unique(dat$id)
    ni  <- c(unlist(tapply(dat$id, dat$id, length)))  
    SampVar <- NULL  
    for(i in uid){ yi      <- dat$Y[dat$id==i]
                   xi      <- dat$X[dat$id==i,1:2]
                   SampVar <- rbind(SampVar, c(mean(yi), solve(t(xi)%*%xi) %*% t(xi) %*% yi))
    }
    SampVar <- (w.function=="mean")*SampVar[,1] + 
               (w.function=="intercept")*SampVar[,2] + 
               (w.function=="slope")*SampVar[,3]

    SampStratum  <- ifelse(SampVar<C1, 1, 
                    ifelse(SampVar<C2, 2, 3))
    NperStratum  <- unlist(tapply(uid, SampStratum, length))
  
    SampProbiThry <- ifelse(SampVar<C1, SampProbThry[1],
                     ifelse(SampVar<C2, SampProbThry[2], SampProbThry[3]))
    
    Sampled     <- rbinom(length(SampProbiThry), 1, SampProbiThry)
    SampProbObs <- c(tapply(Sampled, SampStratum, mean))

    SampProbiObs  <- ifelse(SampVar<C1, SampProbObs[1],
                     ifelse(SampVar<C2, SampProbObs[2], SampProbObs[3]))
    #print(rbind(SampProbThry, SampProbObs))
    #print(cbind(SampProbiThry,SampProbiObs))
    

    ## Independent Sampling
    if (SamplingStrategy=="IndepODS") InODS <- uid[ Sampled==1] 
     
    TheSample <- dat$id %in% InODS                     
    X.ods     <- dat$X[TheSample,]
    Y.ods     <- dat$Y[TheSample]
    Z.ods     <- dat$Z[TheSample,]
    id.ods    <- dat$id[TheSample]
    if (SamplingStrategy=="IndepODS"){
        SampProbi.ods <- rep(SampProbiThry, ni)[TheSample]
        SampProb.ods  <- SampProbThry
    }
    SampStrat.ods <- SampStratum[InODS]
    Qi <- SampVar[InODS]
    dup.id <- duplicated(dat$id)
    dat.univariate <- dat$X[!dup.id,]
    dat.univ.ods   <- dat.univariate[InODS,]

    SampProb  <- SampProbThry
    SampProbi <- rep(SampProbiThry, ni)
    Qi        <- SampVar
     
    list(X=dat$X, Y=dat$Y, Z=dat$Z, id=dat$id,
      SampProb=SampProb, SampProbi=SampProbi,
      N=dat$N, n=dat$n, beta=dat$beta, sig.b0=dat$sig.b0, 
      sig.b1=dat$sig.b1, rho=dat$rho, sig.e=dat$sig.e,
      prob.grp=dat$prob.grp, w.function=w.function,
      cutpoint=c(C1,C2), SampStratum=SampStratum, Qi=Qi, InSample=TheSample)

}
#########################################################
#########################################################
#########################################################
## note that we really do not need quants, PopnQuants, and w.function but to make the fitting function work
Random.Sampling <- function(d, quants=c(.1, .9), PopnQuants, w.function="mean", n=225){
    s <- sample(unique(d$id), n)     
    TheSample <- d$id %in% s
    X.rand     <- d$X
    Y.rand     <- d$Y
    Z.rand     <- d$Z
    id.rand    <- d$id
    
    C1 <- PopnQuants[2,match(quants[1], PopnQuants[1,])]
    C2 <- PopnQuants[2,match(quants[2], PopnQuants[1,])]

    list(X=X.rand, Y=Y.rand, Z=Z.rand, id=id.rand, InSample=TheSample,
         N=d$N, n=d$n, beta=d$beta, sig.b0=d$sig.b0, 
         sig.b1=d$sig.b1, rho=d$rho, sig.e=d$sig.e,
         prob.grp=d$prob.grp,
         ## output below is only needed for the fitting function to run.  They are not used.
         SampProb=c(1,1,1), cutpoint=c(C1,C2), SampProbi=rep(1, length(Y.rand)), w.function=w.function)
}
